"""Compact GA minimal implementation"""

from __future__ import division, print_function
import numpy as np  # to replace 'inf' with np.inf and vice versa


class CompactGA(object):
    """Randomized search on the binary domain {0,1}^n.

    cGA samples two different solutions per iteration.

    Minimal example::

        dimension = 20
        opt = CompactGA(dimension * [0.5])  # initialise in domain middle
        while opt.best.f > 0 and not opt.stop():
            opt.optimize(sum, 1)
        print("%d: fbest=%f" % (opt.evaluation, opt.best.f))

    finds argmin_{x in {0, 1}**20} sum(x) in ``opt.result``.

    Details: due to the incrememental update, cGA is expected to show inferior
    scaling with dimension on LeadingOnes and/or BinVal.

    Reference: Harik et al 1999.
    """

    def __init__(self, mean_):
        """takes as input the initial vector of probabilities to sample 1 vs 0.

        The vector length defines the dimension. Initial values of the
        probability vector, which is also the distribution mean, are
        usually chosen to be 0.5.
        """
        self.mean = np.asarray(mean_)
        self.dimension = len(self.mean)
        self.eta = 1 / self.dimension  # 1 / dim according to Harik et al 1999
        self.lower_p = 1 / self.dimension / 2
        """lower and 1 - upper bound for mean, avoid getting stuck "forever" """
        # bookkeeping
        self.best = BestSolution()
        self.fcurrent = np.nan
        self.evaluation = 0

    def ask(self):
        """return two candidate solutions"""
        x1 = np.random.rand(self.dimension) < self.mean
        x2 = np.random.rand(self.dimension) < self.mean
        while all(x2 == x1):
            x2 = np.random.rand(self.dimension) <= self.mean
        return [np.array(x1, np.int8), np.array(x2, np.int8)]

    def tell(self, solutions, fitness_values):
        """update ``self.mean``"""
        assert len(solutions) == len(fitness_values) == 2
        self.evaluation += 2
        x1, x2 = solutions
        f1, f2 = fitness_values
        if f2 < f1:  # swap such that f1 is better
            f1, f2 = f2, f1
            x1, x2 = x2, x1
        if f1 < f2:  # update mean
            self.mean += self.eta * (x1 - x2)
            self.mean[self.mean < self.lower_p] = self.lower_p
            self.mean[self.mean > 1 - self.lower_p] = 1 - self.lower_p
        # bookkeeping
        self.fcurrent = f1
        self.best.update(f1, x1, self.evaluation)

    def optimize(self, fun, iterations):
        for _ in range(iterations):
            X = self.ask()
            self.tell(X, [fun(X[0]), fun(X[1])])
            if self.stop():
                break
        return self

    def stop(self):
        """dictionary containing satisfied termination conditions
        """
        stop_dict = {}
        self.stop_stagnation_evals = 2000 * self.dimension
        if self.evaluation > self.best.evaluation + self.stop_stagnation_evals:
            stop_dict['stagnation'] = self.stop_stagnation_evals
        return stop_dict

    @property
    def result(self):
        """for the time being `result` is the best seen solution
        """
        return self.best.x

class BestSolution(object):
    """Helper class to keep track of the best solution (minimization).

    All is stored in attributes ``x, f, evaluation``. The only reason
    for this class is to prevent code duplication of the `update`
    method.
    """
    def __init__(self):
        self.x = None
        self.f = np.inf
        self.evaluation = 0
    def update(self, f, x, evaluation):
        """update attributes ``f, x, evaluation`` if ``f < self.f``"""
        if f < self.f:
            self.f = f
            self.x = x[:]
            self.evaluation = evaluation




# -*- coding: utf-8 -*-
"""Package `cma` implements the CMA-ES (Covariance Matrix Adaptation
Evolution Strategy).

CMA-ES is a stochastic optimizer for robust non-linear non-convex
derivative- and function-value-free numerical optimization.

This release was tested with Python versions 3.8 to 3.13. The
implementation is intended to be compatible with Python >= 2.7.

CMA-ES searches for a minimizer (a solution x in :math:`R^n`) of an
objective function f (cost function), such that f(x) is minimal. Regarding
f, only a passably reliable ranking of the candidate solutions in each
iteration is necessary. Neither the function values themselves, nor the
gradient of f need to be available or do matter, like in the downhill
simplex Nelder-Mead algorithm. Some termination criteria however depend on
actual Delta f-values.

The `cma` module provides two independent implementations of the
CMA-ES algorithm in the classes `cma.CMAEvolutionStrategy` and
`cma.purecma.CMAES`.

In each implementation two interfaces are provided:

- functions `fmin2` and `purecma.fmin`:
    run a complete minimization of the passed objective function with
    CMA-ES. `fmin2` also provides optional restarts and noise handling.

- class `CMAEvolutionStrategy` (and the alias `CMA`) and `purecma.CMAES`:
    allow for minimization such that the control of the iteration loop
    remains with the user. `fmin2` returns an instance of
    `CMAEvolutionStrategy`.

Additionally, `fmin_con2` provides constrained optimization.

For a quick start see below or confer to the notebook(s) https://github.com/CMA-ES/pycma/blob/development/notebooks/notebook-usecases-basics.ipynb

`CMAEvolutionStrategy` relies, in contrast to `cma.purecma`, heavily on
`numpy` and optionally on `matplotlib.pyplot` (for `plot` etc., optional
but highly recommended).

The source code is available at https://github.com/CMA-ES/pycma.

Testing
=======
From the system shell::

    python -m cma.test -h
    python -m cma.test
    python -c "import cma.test; cma.test.main()"  # the same

or from any (i)python shell::

    import cma.test
    cma.test.main()

should run without complaints in about between 20 and 100 seconds.

Example
=======
From a python shell::

    import cma
    help(cma)  # "this" help message, use cma? in ipython
    help(cma.fmin)
    help(cma.CMAEvolutionStrategy)
    help(cma.CMAOptions)
    cma.CMAOptions('tol')  # display 'tolerance' termination options
    cma.CMAOptions('verb') # display verbosity options
    x, es = cma.fmin2(cma.ff.tablet, 15 * [1], 1)
    es = cma.CMAEvolutionStrategy(15 * [1], 1).optimize(cma.ff.tablet)
    help(es.result)
    x, es.result[0]  # best evaluated solution
    es.result[5]  # mean solution, presumably better with noise

:See also: `fmin` (), `CMAOptions`, `CMAEvolutionStrategy`

:Author: Nikolaus Hansen, 2008-
:Author: Petr Baudis, 2014
:Author: Youhei Akimoto, 2017-

:License: BSD 3-Clause, see LICENSE file.

"""

# How to create a html documentation file (see also setup-howto.py):
#    pydoctor --docformat=restructuredtext --html-output=apidocs cma > pydoctor-messages.txt
# old:
#    pydoctor --docformat=restructuredtext --make-html cma
#    pydoc -w cma  # edit the header (remove local pointers)
#    epydoc cma.py  # comes close to javadoc but does not find the
#                   # links of function references etc
#    doxygen needs @package cma as first line in the module docstring
#       some things like class attributes are not interpreted correctly
#    sphinx: doc style of doc.python.org, could not make it work (yet)
# __docformat__ = "reStructuredText"  # this hides some comments entirely?

from __future__ import absolute_import as _ab # now local imports must use .
from __future__ import division as _di
from __future__ import print_function as _pr
del _ab, _di, _pr


___author__ = "Nikolaus Hansen and Petr Baudis and Youhei Akimoto"
__license__ = "BSD 3-clause"
__version__ = "4.4.4"
'''version 4.4.4'''  # allows the apidocs to "pick up" the version number


import collections as _collections
import warnings as _warnings

# __package__ = 'cma'
from .more_algorithms import purecma
try:
    import numpy as _np
except ImportError:
    _warnings.warn('Only `cma.purecma` has been imported. Install `numpy` ("pip'
                   ' install numpy") if you want to import the entire `cma`'
                   ' package.')
else:
    from . import (constraints_handler, evolution_strategy, fitness_functions,
                    fitness_transformations, interfaces, optimization_tools,
                    sampler, sigma_adaptation, transformations, utilities,
                    )
    # from . import test  # gives a warning with python -m cma.test (since Python 3.5.3?)
    test = 'type "import cma.test" to access the `test` module of `cma`'
    from . import s
    from .fitness_functions import ff
    from .fitness_transformations import GlueArguments, ScaleCoordinates
    from .evolution_strategy import fmin, fmin2, fmin_con, fmin_con2, fmin_lq_surr, fmin_lq_surr2
    from .evolution_strategy import CMAEvolutionStrategy
    from .options_parameters import CMAOptions, cma_default_options_
    CMA = CMAEvolutionStrategy  # shortcut for typing without completion
    from .logger import disp, plot, plot_zip, CMADataLogger
    from .optimization_tools import NoiseHandler
    from .boundary_handler import BoundPenalty, BoundTransform, BoundNone, BoundDomainTransform
    from .constraints_handler import ConstrainedFitnessAL, AugmentedLagrangian
    constraints_handler.BoundTransform = BoundTransform  # for backwards compatibility

try:
    version_info = _collections.namedtuple(
            'version_info', ('major', 'minor', 'micro'), module='cma')(
                    *(int(s) for s in __version__.split(' ')[0].split('.')))
except TypeError:
    try:
        version_info = _collections.namedtuple(
                'version_info', ('major', 'minor', 'micro'))(
                        *(int(s) for s in __version__.split(' ')[0].split('.')))
    except Exception:
        version_info = tuple(int(s) for s in __version__.split(' ')[0].split('.'))
# a simple tuple is in the end more user friendly!?
version_info = tuple(int(s) for s in __version__.split(' ')[0].split('.'))




"""versatile shortcuts for quick typing in an (i)python shell,

 or even using ``from cma.s import *`` in interactive sessions.

Provides various aliases from within the `cma` package, to be reached like
``cma.s....``

Do not use this module for stable code developments!

This is not actively maintained.
"""
import warnings as _warnings
try: from matplotlib import pyplot as _pyplot  # like this it doesn't show up in the interface
except ImportError:
    _pyplot = None
    _warnings.warn('Could not import matplotlib.pyplot, therefore'
                   ' ``cma.plot()`` etc. is not available')
# from . import fitness_functions as ff
from . import evolution_strategy as es
from . import fitness_transformations as ft
from . import transformations as tf
from . import constraints_handler as ch
from .utilities import utils
from .evolution_strategy import CMAEvolutionStrategy as _CMAES
from .utilities.utils import pprint
from .utilities.math import Mh, testchisquare, testranksum
from .utilities.utils import figure
# from .fitness_functions import elli as felli

class CMAES(_CMAES):
    def __init__(self, *args, **kwargs):
        _warnings.warn("Deprecated (renamed). Use `cma.CMA` instead of `cma.s.CMAES`",
                       FutureWarning)
        super(CMAES, self).__init__(*args, **kwargs)

if _pyplot:
    def figshow():
        """`pyplot.show` to make a plotted figure show up"""
        # is_interactive = matplotlib.is_interactive()

        _pyplot.ion()
        _pyplot.show()
        # if we call now matplotlib.interactive(True), the console is
        # blocked
    figsave = _pyplot.savefig

if 11 < 3:
    try:
        from matplotlib.pyplot import savefig as figsave, close as figclose, ion as figion
    except:
        figsave, figclose, figshow = 3 * ['not available']
        _warnings.warn('Could not import matplotlib.pyplot, therefore ``cma.plot()``'
                       ' etc. is not available')

def _cdict(obj, exclude='_'):
    """return public attributes of `obj` as a `dict`.

    Pass '__' as second argument to see "private" attributes.
    """
    return {d[0]:d[1] for d in obj.__dict__.items() if not d[0].startswith(exclude)}
def clean(*args, **kwargs):
    """Deprecated, renamed to `ddir`"""
    _warnings.warn("Deprecated (renamed). Use `ddir` instead of `clean`", FutureWarning)
    return ddir(*args, **kwargs)
def ddir(obj, exclude='_'):
    """return "public" elements of a `list` or `dict` or of ``object.__dict__``.

    Ignore entries starting with `exclude`. Return a 'list` when `obj`
    is a `list` or a `dict`, return a `dict` otherwise. This is versatile
    and could change in future.

    Typical usage: ``clean(dir())`` or ``clean(obj)`` or ``clean(dir(), '__')``.
    """
    if isinstance(obj, (list, dict)):
        return [d for d in obj if not d.startswith(exclude)]
    return _cdict(obj, exclude=exclude)



# -*- coding: utf-8 -*-
"""A collection of variable boundaries (AKA box constraints) handling classes.
"""
from __future__ import absolute_import, division, print_function  #, unicode_literals
import warnings as _warnings
import numpy as np
from .utilities import utils as _utils
from .utilities.utils import rglen
from .transformations import BoxConstraintsLinQuadTransformation
del absolute_import, division, print_function  #, unicode_literals

def normalize_bounds(bounds, copy=True):
    """return ``[lower_bounds, upper_bounds]`` such that each of them

    is either a nonempty `list` or `None`.

    `copy` only makes a shallow copy.

    On input, `bounds` must be a `list` of length 2.
    """
    for i in [0, 1]:
        try:
            if len(bounds[i]) == 0:  # bails when bounds[i] is a scalar
                if copy:
                    bounds = list(bounds)
                    copy = False
                bounds[i] = None  # let's use None instead of empty list
        except TypeError:
            if copy:
                bounds = list(bounds)
                copy = False
            bounds[i] = [bounds[i]]
        if not _utils.is_(bounds[i]) or all(
                bounds[i][j] is None or not np.isfinite(bounds[i][j])
                    for j in rglen(bounds[i])):
            if copy:
                bounds = list(bounds)
                copy = False
            bounds[i] = None
        if bounds[i] is not None and any(bounds[i][j] == (-1)**i * np.inf
                                            for j in rglen(bounds[i])):
            raise ValueError('lower/upper is +inf/-inf and ' +
                                'therefore no finite feasible solution is available')
    return bounds

def none_to_inf(bounds, copy=True):
    """replace any None by inf or -inf.

    This code was never tested and is not in use.
    """
    copy_i = [copy, copy]
    for ib, val in [[0, -np.inf], [1, np.inf]]:
        if bounds[ib] is None:
            if copy:
                bounds = list(bounds)
                copy = False
            bounds[ib] = [val]
            continue
        for i in rglen(bounds[ib]):
            if bounds[ib][i] is None:
                if copy_i[ib]:
                    bounds[ib] = list(bounds[ib])
                    copy_i[ib] = False
                bounds[ib][i] = val
    return bounds

class BoundDomainTransform(object):
    """create a `callable` with unbounded domain from a function with bounded domain,

    for example an objective or constraints function. The new "unbounded"
    `callable` maps ``x`` to ``function(self.transform(x))``, first
    "projecting" the input (e.g., candidate solutions) into the bounded
    domain with `transform` before to evaluate the result on the original
    function. The "projection" is smooth and differentiable, namely
    coordinate-wise piecewise linear or quadratic. The "projection" is not
    idempotent.

    Bounds are passed as ``boundaries = [lower_bounds, upper_bounds]``
    where each ``*_bounds`` can be `None`, or a scalar, or an iterable (a
    vector) whose values can be `None` too. If the iterable has fewer
    values than the solutions, the last value is recycled, if it has more
    values, trailing values are ignored.

    For example, ``boundaries = [None, [0, 0, None]]`` means no lower
    bounds and an upper bound of zero for the first two variables.

    Example:

    >>> import cma
    >>> fun = cma.boundary_handler.BoundDomainTransform(
    ...             cma.ff.sphere,  # is composed with fun.transform into fun
    ...             [[0.02, 0.01], None])  # boundaries for the "original" problem
    >>> x, es = cma.fmin2(fun, 3 * [0.5], 0.5, {'verbose':-9})
    >>> assert all(x - 1e-4 < [-0.03, -0.04, -0.04])  # x is in the unbounded domain
    >>> print("solution in the original (bounded) domain = {}"
    ...       .format(fun.transform(es.result.xfavorite)))
    solution in the original (bounded) domain = [0.02 0.01 0.01]

    The original function can be accessed and called via the attribute
    ``function`` like ``fun.function(...)``. For code simplicity,
    attributes of ``function`` are "inherited" to ``fun``.

    Details: the resulting function has a repetitive landscape with a
    period slightly larger than twice the boundary interval. The
    ``BoundDomainTransform(function,...)`` instance emulates the original function
    on attribute access by calling ``__getattr__``. To access shadowed
    attributes or for debugging, replace ``.attrname`` with
    ``.function.attrname``.
    """
    def __init__(self, function, boundaries):
        """return a callable that evaluates `function` only within `boundaries`"""
        self.function = function
        self.boundary_handler = BoundTransform(boundaries)
        self.transform = self.boundary_handler.transform  # alias for convenience
    def __call__(self, x, *args, **kwargs):
        return self.function(self.transform(x), *args, **kwargs)
    def __getattr__(self, name):
        """return ``getattr(self.function, name)`` when ``not hasattr(self, name)``.

        This emulates the `function` interface, kinda like blind inheritance.
        """
        return getattr(self.function, name)

class _BoundDomainPenalty(object):
    """[WIP early] function wrapper for penalty boundary handler, looks kinda complex"""
    def __init__(self, function, boundaries):
        """return a callable that evaluates `function` only within `boundaries`"""
        self.function = function
        self.boundary_handler = BoundPenalty(boundaries)
        self.transform = self.boundary_handler.repair  # alias for convenience
    def __call__(self, x, *args, **kwargs):
        return self.function(self.transform(x), *args, **kwargs)
    def callback(self):
        """needs the mean and the fitness values?"""
        raise NotImplementedError
    def __getattr__(self, name):
        """return ``getattr(self.function, name)`` when ``not hasattr(self, name)``.

        This emulates the `function` interface, kinda like blind inheritance.
        """
        return getattr(self.function, name)

class BoundaryHandlerBase(object):
    """quick hack versatile base class.

    To guaranty that modifications of the attribute `bounds` after this
    instance was already used are effective, the class attribute
    ``_bounds_dict`` must be reset like ``_bounds_dict = {}``.
    """
    use_cached_values = False
    '''default behavior as to whether to use cached values'''
    def __init__(self, bounds):
        """`bounds` can be ``None`` or ``[lb, ub]``

        where ``lb`` and ``ub`` are either ``None`` or a vector (which can have
        ``None`` entries).

        On return, the ``bounds`` attribute of ``self`` are the bounds in a
        normalized form.

        To compute bounds for any dimension, the last entry of bounds is
        then recycled for variables with ``indices >= len(bounds[i]) for i
        in (0,1)``.

        """
        self.use_cached_values = BoundaryHandlerBase.use_cached_values
        if bounds in [None, (), []]:
            self.bounds = None
        else:
            if not isinstance(bounds, (tuple, list)) or len(bounds) != 2:
                raise ValueError(
                    "bounds must be None, empty, or a list of length 2"
                    " where each element may be a scalar, list, array,"
                    " or None; type(bounds) was: %s" % str(type(bounds)))
            self.bounds = normalize_bounds(list(bounds), copy=False)
        self._bounds_dict = {}
        '''saved return values of get_bounds(i, dim). Changing `self.bounds` may
           only be effective when this is reset.'''

    def amend_bounds_for_integer_variables(self, integer_indices, at=0.5, offset=1e-9):
        """set bounds away from ``at=0.5`` such that

        a repaired solution is always rounded into the feasible domain.
        """
        if not integer_indices or not self.has_bounds():
            return
        def set_bounds(which, idx, integer_indices):
            """idx is a `bool` index array of ``bound % 1 == 0.5``"""
            assert which in (0, 1), which
            if not np.any(idx):
                return
            for i in np.nonzero(idx)[0]:
                if i not in integer_indices:
                    idx[i] = False
            if not np.any(idx):
                return
            dimension = max((len(self.bounds[which]),
                             np.max(np.nonzero(idx)[0]) + 1))
            bounds = self.get_bounds(which, dimension)
            if len(bounds) < len(idx):
                idx = idx[:len(bounds)]  # last nonzero entry determined len
            elif len(bounds) > len(idx):
                idx = np.hstack([idx, np.zeros(len(bounds) - len(idx), dtype=bool)])
            bounds[idx] += (1 - 2 * which) * offset * np.maximum(1, np.abs(bounds[idx]))
            self.bounds[which] = bounds
            self._bounds_dict = {}

        dimension = max(integer_indices) + 1
        for which in (0, 1):
            if not self.has_bounds('upper' if which else 'lower'):
                continue
            bounds = self.get_bounds(which, dimension)
            set_bounds(which, np.mod(bounds, 1) == at, integer_indices)

    def __call__(self, solutions, *args, **kwargs):
        """return penalty or list of penalties, by default zero(s).

        This interface seems too specifically tailored to the derived
        BoundPenalty class, it should maybe change.

        """
        return 0.0 if np.isscalar(solutions[0]) else len(solutions) * [0.0]

    def update(self, *args, **kwargs):
        """end-iteration callback of boundary handler (abstract/empty)"""
        return self

    def repair(self, x, copy_if_changed=True):
        """projects infeasible values on the domain bound, might be
        overwritten by derived class """
        copy = copy_if_changed
        if self.bounds is None:
            return x
        for ib in [0, 1]:
            if self.bounds[ib] is None:
                continue
            for i in rglen(x):
                idx = min([i, len(self.bounds[ib]) - 1])
                if self.bounds[ib][idx] is not None and \
                        (-1)**ib * x[i] < (-1)**ib * self.bounds[ib][idx]:
                    if copy:
                        x = np.array(x, copy=True)
                        copy = False
                    x[i] = self.bounds[ib][idx]
        return x

    def inverse(self, y, copy_if_changed=True):
        """inverse of repair if it exists, at least it should hold
         ``repair == repair o inverse o repair``"""
        return y

    def get_bounds(self, which, dimension):
        """``get_bounds('lower', 8)`` returns the lower bounds in 8-D"""
        if which in ['lower', 0, '0']:
            return self._get_bounds(0, dimension)
        elif which in ['upper', 1, '1']:
            return self._get_bounds(1, dimension)
        else:
            raise ValueError("argument which must be 'lower' or 'upper'")

    def _get_bounds(self, ib, dimension):
        """ib == 0/1 means lower/upper bound, return a vector of length
        `dimension` """
        if not self.use_cached_values:
            self._bounds_dict = {}  # recover original behavior
        if (ib, dimension) in self._bounds_dict:
            return self._bounds_dict[(ib, dimension)]
        sign_ = 2 * ib - 1
        assert sign_**2 == 1
        if self.bounds is None or self.bounds[ib] is None:
            self._bounds_dict[(ib, dimension)] = sign_ * np.inf + np.zeros(dimension)
            return self._bounds_dict[(ib, dimension)]
        res = []
        for i in range(dimension):
            res.append(self.bounds[ib][min([i, len(self.bounds[ib]) - 1])])
            if res[-1] is None:
                res[-1] = sign_ * np.inf
        self._bounds_dict[(ib, dimension)] = np.asarray(res)
        return self._bounds_dict[(ib, dimension)]

    def has_bounds(self, which='both'):
        """return `True` if any variable is bounded"""
        valid_whichs = (None, 'both', 'lower', 'upper')
        if which not in valid_whichs:
            raise ValueError("`which` parameter must be in {0} but was ={1}"
                                .format(valid_whichs, which))
        bounds = self.bounds
        if bounds is None or bounds in (False, [], ()) or (
                bounds[0] is None and bounds[1] is None):
            return False
        for ib, bound in enumerate(bounds):
            if bound is None or (
                ib == 0 and which == 'upper') or (
                ib == 1 and which == 'lower'):
                continue
            sign_ = 2 * ib - 1
            for bound_i in bound:
                if bound_i is not None and sign_ * bound_i < np.inf:
                    return True
        return False

    def is_in_bounds(self, x):
        """return `True` if `x` is in bounds.

        >>> import numpy as np
        >>> import cma

        >>> b = cma.boundary_handler.BoundaryHandlerBase(bounds=[[-0.5], 0.2])
        >>> for n in [2, 4, 9]:
        ...     x = np.random.randn(n)
        ...     assert (all(-0.5 <= x) and all(0.2 >= x)) is b.is_in_bounds(x)

        """
        if self.bounds is None:
            return True
        for ib in [0, 1]:
            if self.bounds[ib] is None:
                continue
            for i in rglen(x):
                if i == len(self.bounds[ib]) - 1 and (
                        self.bounds[ib][i] is None or
                        (-1)**(1 - ib) * self.bounds[ib][i] == np.inf):
                    break
                idx = min([i, len(self.bounds[ib]) - 1])
                if self.bounds[ib][idx] is None:
                    continue
                if ((ib == 0 and x[i] < self.bounds[ib][idx]) or (
                     ib == 1 and x[i] > self.bounds[ib][idx])):
                    return False
        return True

    def into_bounds(self, x, copy=True):
        """set out-of-bound values on bounds and return `x`.

        Make a copy when `x` is changed and ``copy is True``,
        otherwise change in place.

        >>> import numpy as np
        >>> import cma

        >>> b = cma.boundary_handler.BoundaryHandlerBase(bounds=[[0.1], 0.2])
        >>> assert all(0.1 <= b.into_bounds(np.random.randn(22)))
        >>> assert all(0.2 >= b.into_bounds(np.random.randn(11)))
        >>> b = cma.boundary_handler.BoundaryHandlerBase(bounds=[[-0.1], np.inf])
        >>> assert all(-0.1 <= b.into_bounds(np.random.randn(22)))
        >>> b = cma.boundary_handler.BoundaryHandlerBase(bounds=[[-0.1], [None]])
        >>> assert all(-0.1 <= b.into_bounds(np.random.randn(22)))
        >>> b = cma.boundary_handler.BoundaryHandlerBase(bounds=[-np.inf, 0.1])
        >>> assert all(0.1 >= b.into_bounds(np.random.randn(22)))

        """
        if self.bounds is None:
            return x
        for ib in [0, 1]:
            if self.bounds[ib] is None:
                continue
            for i in rglen(x):
                if i == len(self.bounds[ib]) - 1 and (
                          self.bounds[ib][i] is None or
                          (-1)**(1 - ib) * self.bounds[ib][i] == np.inf):
                    break
                idx = min([i, len(self.bounds[ib]) - 1])
                if self.bounds[ib][idx] is None:
                    continue
                if ((ib == 0 and x[i] >= self.bounds[ib][idx]) or (
                     ib == 1 and x[i] <= self.bounds[ib][idx])):
                    continue
                if copy:
                    x = np.array(x, copy=True)
                    copy = False  # copy only once
                x[i] = self.bounds[ib][idx]
        return x

    def idx_out_of_bounds(self, x):
        """return index list of out-of-bound values in `x`.

        ``if bounds.idx_out_of_bounds`` evaluates to `True` if and only if
        `x` is out of bounds.
        """
        if self.bounds is None:
            return []
        idxs = []
        for ib in [0, 1]:
            if self.bounds[ib] is None:
                continue
            for i in rglen(x):
                idx = min([i, len(self.bounds[ib]) - 1])
                if self.bounds[ib][idx] is None:
                    continue
                if ((ib == 0 and x[i] < self.bounds[ib][idx]) or (
                     ib == 1 and x[i] > self.bounds[ib][idx])):
                   idxs += [i]
        return sorted(idxs)

    def get_bound(self, index):
        """return lower and upper bound of variable with index `index`"""
        if self.bounds is None:
            return [-np.inf, np.inf]
        res = []
        for ib in [0, 1]:
            if self.bounds[ib] is None or len(self.bounds[ib]) == 0:
                b = None
            else:
                b = self.bounds[ib][min((index, len(self.bounds[ib]) - 1))]
            res.append([-np.inf, np.inf][ib] if b is None else b)
        return res

    def to_dim_times_two(self, bounds=None):
        """return boundaries in format ``[[lb0, ub0], [lb1, ub1], ...]``,
        as used by ``BoxConstraints...`` class.

        Use by default ``bounds = self.bounds``.
        """
        if bounds is None:
            try:
                bounds = self.bounds
            except AttributeError:
                pass
        if not bounds:
            return [[None, None]]
        bounds = normalize_bounds(list(bounds), copy=False)
        for i in [0, 1]:
            if bounds[i] is None:  # replace None with [None] which is easier below
                bounds[i] = [None]  # [-np.inf if i == 0 else np.inf]
        l = [len(bounds[i]) for i in [0, 1]]
        if l[0] != l[1] and 1 not in l and None not in (
                bounds[0][-1], bounds[1][-1]):  # warn on different lengths
            _warnings.warn(
                "lower and upper bounds do not have the same length or length\n"
                " one or `None` as last element (the last"
                    " element is always recycled).\n"
                "Lengths are {0} = [len(b) for b in bounds={1}])".format(l, bounds))
        b = []  # bounds in different format
        try:
            for i in range(max(l)):
                b.append([bounds[0][min((i, l[0] - 1))],
                            bounds[1][min((i, l[1] - 1))]])
        except (TypeError, IndexError):
            print("boundaries must be provided in the form " +
                    "[scalar_of_vector, scalar_or_vector]")
            raise
        return b

class BoundNone(BoundaryHandlerBase):
    """no boundaries"""
    def __init__(self, bounds=None):
        if bounds is not None:
            raise ValueError()
        # BoundaryHandlerBase.__init__(self, None)
        super(BoundNone, self).__init__(None)
    def is_in_bounds(self, x):
        return True

class BoundTransform(BoundaryHandlerBase):
    """Handle boundaries by a smooth, piecewise linear and quadratic
    transformation into the feasible domain.

    >>> import numpy as np
    >>> import cma
    >>> from cma.boundary_handler import BoundTransform
    >>> from cma import fitness_transformations as ft
    >>> veq = cma.utilities.math.Mh.vequals_approximately
    >>> b = BoundTransform([0, None])
    >>> assert b.bounds == [[0], None]
    >>> assert veq(b.repair([-0.1, 0, 1, 1.2]), np.array([0.0125, 0.0125, 1, 1.2])), b.repair([-0.1, 0, 1, 1.2])
    >>> assert b.is_in_bounds([0, 0.5, 1])
    >>> assert veq(b.transform([-1, 0, 1, 2]), [0.9, 0.0125,  1,  2  ]), b.transform([-1, 0, 1, 2])
    >>> bounded_sphere = ft.ComposedFunction([
    ...         cma.ff.sphere,
    ...         BoundTransform([[], 5 * [-1] + [np.inf]]).transform
    ...     ])
    >>> o1 = cma.fmin(bounded_sphere, 6 * [-2], 0.5)  # doctest: +ELLIPSIS
    (4_w,9)-aCMA-ES (mu_w=2.8,w_1=49%) in dimension 6 (seed=...
    >>> o2 = cma.fmin(cma.ff.sphere, 6 * [-2], 0.5, options={
    ...    'BoundaryHandler': cma.boundary_handler.BoundTransform,
    ...    'bounds': [[], 5 * [-1] + [np.inf]] })  # doctest: +ELLIPSIS
    (4_w,9)-aCMA-ES (mu_w=2.8,w_1=49%) in dimension 6 (seed=...
    >>> assert o1[1] < 5 + 1e-8 and o2[1] < 5 + 1e-8
    >>> b = BoundTransform([-np.random.rand(120), np.random.rand(120)])
    >>> for i in range(0, 100, 9):
    ...     x = (-i-1) * np.random.rand(120) + i * np.random.randn(120)
    ...     x_to_b = b.repair(x)
    ...     x2 = b.inverse(x_to_b)
    ...     x2_to_b = b.repair(x2)
    ...     x3 = b.inverse(x2_to_b)
    ...     x3_to_b = b.repair(x3)
    ...     assert veq(x_to_b, x2_to_b)
    ...     assert veq(x2, x3)
    ...     assert veq(x2_to_b, x3_to_b)
    >>> for _ in range(5):
    ...     lb = np.random.randn(4)
    ...     ub = lb + 1e-7 + np.random.rand(4)
    ...     b = BoundTransform([lb, ub])
    ...     for x in [np.random.randn(4) / np.sqrt(np.random.rand(4)) for _ in range(22)]:
    ...         assert all(lb <= b.transform(x)), (lb, ub, b.__dict__)
    ...         assert all(b.transform(x) <= ub), (lb, ub, b.__dict__)

    Details: this class uses ``class BoxConstraintsLinQuadTransformation``

    """
    def __init__(self, bounds=None):
        """Argument bounds can be `None` or ``bounds[0]`` and ``bounds[1]``
        are lower and upper domain boundaries, each is either `None` or
        a scalar or a list or array of appropriate size.

        """
        # BoundaryHandlerBase.__init__(self, bounds)
        super(BoundTransform, self).__init__(bounds)
        self.bounds_tf = BoxConstraintsLinQuadTransformation(self.to_dim_times_two(bounds))

    def repair(self, x, copy_if_changed=True):
        """transforms ``x`` into the bounded domain.
        """
        copy = copy_if_changed
        if self.bounds is None or (self.bounds[0] is None and
                                   self.bounds[1] is None):
            return x
        return np.asarray(self.bounds_tf(x, copy))

    def transform(self, x):
        return self.repair(x)

    def inverse(self, x, copy_if_changed=True):
        """inverse transform of ``x`` from the bounded domain.

        """
        if self.bounds is None or (self.bounds[0] is None and
                                   self.bounds[1] is None):
            return x
        return np.asarray(self.bounds_tf.inverse(x, copy_if_changed))  # this doesn't exist

class BoundPenalty(BoundaryHandlerBase):
    """Compute a bound penalty and update coordinate-wise penalty weights.

    An instance must be updated each iteration using the `update` method.

    Details:

    - The penalty computes like ``sum(w[i] * (x[i]-xfeas[i])**2)``,
      where ``xfeas`` is the closest feasible (in-bounds) solution from
      ``x``. The weight ``w[i]`` should be updated during each iteration
      using the update method.

    Example how this boundary handler is used with `cma.fmin` via the
    options (`CMAOptions`) of the class `cma.CMAEvolutionStrategy`:

    >>> import cma
    >>> res = cma.fmin(cma.ff.elli, 6 * [0.9], 0.1,
    ...     {'BoundaryHandler': cma.BoundPenalty,
    ...      'bounds': [-1, 1],
    ...      'tolflatfitness': 10,
    ...      'fixed_variables': {0: 0.012, 2:0.234}
    ...     })  # doctest: +ELLIPSIS
    (4_w,8)-aCMA-ES (mu_w=2.6,w_1=52%) in dimension 4 (seed=...
    >>> if res[1] >= 13.76: print(res)  # should never happen

    Reference: Hansen et al 2009, A Method for Handling Uncertainty...
    IEEE TEC, with addendum, see
    https://ieeexplore.ieee.org/abstract/document/4634579
    https://hal.inria.fr/inria-00276216/file/TEC2008.pdf

    **todo**: implement a more generic interface, where this becomes a
    fitness wrapper which adds the desired penalty and the `update`
    method is used as callback argument for `fmin` like::

        f = cma.BoundPenalty(cma.ff.elli, bounds=[-1, 1])
        res = cma.fmin(f, 6 * [1], callback=f.update)

    where callback functions should receive the same arguments as
    `tell`, namely an `CMAEvolutionStrategy` instance, an array of the
    current solutions and their respective f-values. Such change is
    relatively involved. Consider also that bounds are related with the
    geno- to phenotype transformation.
    """
    def __init__(self, bounds=None):
        """Argument bounds can be `None` or ``bounds[0]`` and ``bounds[1]``
        are lower  and upper domain boundaries, each is either `None` or
        a scalar or a `list` or `np.array` of appropriate size.
        """
        # #
        # bounds attribute reminds the domain boundary values
        # BoundaryHandlerBase.__init__(self, bounds)
        super(BoundPenalty, self).__init__(bounds)

        self.gamma = 1  # a very crude assumption
        self.weights_initialized = False  # gamma becomes a vector after initialization
        self.hist = []  # delta-f history

    def repair(self, x, copy_if_changed=True):
        """sets out-of-bounds components of ``x`` on the bounds.

        TODO: unify with BoundaryHandlerBase.repair
        """
        # TODO (old data): CPU(N,lam,iter=20,200,100): 3.3s of 8s for two bounds, 1.8s of 6.5s for one bound
        # remark: np.max([bounds[0], x]) is about 40 times slower than max((bounds[0], x))
        copy = copy_if_changed
        if self.has_bounds():
            bounds = self.bounds
            if copy:
                x = np.array(x, copy=True)
            if bounds[0] is not None:
                if np.isscalar(bounds[0]):
                    for i in rglen(x):
                        x[i] = max((bounds[0], x[i]))
                else:
                    for i in rglen(x):
                        j = min([i, len(bounds[0]) - 1])
                        if bounds[0][j] is not None:
                            x[i] = max((bounds[0][j], x[i]))
            if bounds[1] is not None:
                if np.isscalar(bounds[1]):
                    for i in rglen(x):
                        x[i] = min((bounds[1], x[i]))
                else:
                    for i in rglen(x):
                        j = min((i, len(bounds[1]) - 1))
                        if bounds[1][j] is not None:
                            x[i] = min((bounds[1][j], x[i]))
        return x

    # ____________________________________________________________
    #
    def __call__(self, x, archive, gp):
        """returns the boundary violation penalty for `x`,
        where `x` is a single solution or a list or np.array of solutions.

        """
        # if x in (None, (), []):  # breaks when x is a nparray
        #     return x
        if not self.has_bounds():
            return 0.0 if np.isscalar(x[0]) else [0.0] * len(x)  # no penalty

        x_is_single_vector = np.isscalar(x[0])
        if x_is_single_vector:
            x = [x]

        # add fixed variables to self.gamma
        try:
            gamma = list(self.gamma)  # fails if self.gamma is a scalar
            for i in sorted(gp.fixed_values):  # fails if fixed_values is None
                gamma.insert(i, 0.0)
            gamma = np.asarray(gamma)
        except TypeError:
            gamma = self.gamma
        pen = []
        for xi in x:
            # CAVE: this does not work with already repaired values!!
            # CPU(N,lam,iter=20,200,100)?: 3s of 10s, np.array(xi): 1s
            # remark: one deep copy can be prevented by xold = xi first
            xpheno = xi if gp.isidentity else gp.pheno(archive[xi]['geno'])
            # necessary, because xi was repaired to be in bounds
            xinbounds = self.repair(xpheno)
            # could be omitted (with unpredictable effect in case of external repair)
            fac = 1  # exp(0.1 * (log(self.scal) - np.mean(self.scal)))
            pen.append(sum(gamma * ((xinbounds - xpheno) / fac)**2) / len(xi))
        return pen[0] if x_is_single_vector else pen

    # ____________________________________________________________
    #
    def feasible_ratio(self, solutions):
        """counts for each coordinate the number of feasible values in
        ``solutions`` and returns an `np.array` of length
        ``len(solutions[0])`` with the ratios.
        """
        raise NotImplementedError

    # ____________________________________________________________
    #
    def update(self, function_values, es):
        """updates the weights for computing a boundary penalty.

        Arguments
        =========
        ``function_values``:
            all function values of recent population of solutions
        ``es``:
            `CMAEvolutionStrategy` object instance, in particular
            mean and variances and the methods from the attribute
            `gp` of type `GenoPheno` are used.

        """
        if self.bounds is None or (self.bounds[0] is None and
                                   self.bounds[1] is None):
            return self

        N = es.N
        # ## prepare
        # compute varis = sigma**2 * C_ii
        if 11 < 3:  # old
            varis = es.sigma**2 * np.array(N * [es.C] if np.isscalar(es.C) else (# scalar case
                                    es.C if np.isscalar(es.C[0]) else  # diagonal matrix case
                                    [es.C[i][i] for i in range(N)]))  # full matrix case
        else:
            varis = es.sigma**2 * es.sm.variances

        # relative violation in geno-space
        dmean = (es.mean - es.gp.geno(self.repair(es.gp.pheno(es.mean)))) / varis**0.5

        # ## Store/update a history of delta fitness value
        fvals = sorted(function_values)
        _l = 1 + len(fvals)
        val = fvals[3 * _l // 4] - fvals[_l // 4]  # exact interquartile range apart interpolation
        val = val / np.mean(varis)  # new: val is normalized with sigma of the same iteration
        # insert val in history
        if np.isfinite(val) and val > 0:
            self.hist.insert(0, val)
        elif val == np.inf and len(self.hist) > 1:
            self.hist.insert(0, max(self.hist))
        else:
            pass  # ignore 0 or nan values
        if len(self.hist) > 20 + (3 * N) / es.popsize:
            self.hist.pop()

        # ## prepare
        dfit = np.median(self.hist)  # median interquartile range
        damp = min(1, es.sp.weights.mueff / 10. / N)

        # ## set/update weights
        # Throw initialization error
        if len(self.hist) == 0:
            raise ValueError('wrongful initialization, no feasible solution sampled. ' +
                'Reasons can be mistakenly set bounds (lower bound not smaller than upper bound) or a too large initial sigma0 or... ' +
                'See description of argument func in help(cma.fmin) or an example handling infeasible solutions in help(cma.CMAEvolutionStrategy). ')
        # initialize weights
        if dmean.any() and (not self.weights_initialized or es.countiter == 2):  # TODO
            self.gamma = np.array(N * [2 * dfit])  ## BUGBUGzzzz: N should be phenotypic (bounds are in phenotype), but is genotypic
            self.weights_initialized = True
        # update weights gamma
        if self.weights_initialized:
            edist = np.array(abs(dmean) - 3 * max(1, N**0.5 / es.sp.weights.mueff))
            if 1 < 3:  # this is better, around a factor of two
                # increase single weights possibly with a faster rate than they can decrease
                #     value unit of edst is std dev, 3==random walk of 9 steps
                self.gamma *= np.exp((edist > 0) * np.tanh(edist / 3) / 2.)**damp
                # decrease all weights up to the same level to avoid single extremely small weights
                #    use a constant factor for pseudo-keeping invariance
                self.gamma[self.gamma > 5 * dfit] *= np.exp(-1. / 3)**damp
                #     self.gamma[idx] *= exp(5*dfit/self.gamma[idx] - 1)**(damp/3)
            elif 1 < 3 and (edist > 0).any():  # previous method
                # CAVE: min was max in TEC 2009
                self.gamma[edist > 0] *= 1.1**min(1, es.sp.weights.mueff / 10. / N)
                # max fails on cigtab(N=12,bounds=[0.1,None]):
                # self.gamma[edist>0] *= 1.1**max(1, es.sp.weights.mueff/10./N) # this was a bug!?
                # self.gamma *= exp((edist>0) * np.tanh(edist))**min(1, es.sp.weights.mueff/10./N)
            else:  # alternative version, but not better
                solutions = es.pop  # this has not been checked
                r = self.feasible_ratio(solutions)  # has to be the averaged over N iterations
                self.gamma *= np.exp(np.max([N * [0], 0.3 - r], axis=0))**min(1, es.sp.weights.mueff / 10 / N)
        es.more_to_write += list(self.gamma) if self.weights_initialized else N * [1.0]
        # ## return penalty
        # es.more_to_write = self.gamma if not np.isscalar(self.gamma) else N*[1]
        return self  # bound penalty values




# -*- coding: utf-8 -*-
"""Integer handling class `IntegerCentering` to be used in combination with CMA-ES.

Reference: Marty et al, LB+IC-CMA-ES: Two simple modiﬁcations of CMA-ES to
handle mixed-integer problems. PPSN 2024.
"""
import warnings as _warnings
import numpy as np

centering_on = True
'''convenience flag to turn centering off, read on instantiation of `IntegerCentering`'''

class IntegerCentering(object):
    """round values of int-variables that are different from the int-mean.

    This callable class changes a population of solutions in place, in
    particular rounding some of the integer variables. The call does not
    guaranty to round any or all integer variables. The class assumes that
    the fitness function only interprets the rounded values of integer
    variables, that is, it assumes that int-variables change the fitness
    only if their rounded (genotype) value change. (This may easily become
    a wrong assumption if a geno-pheno transformation is applied). The
    class tries to correct for the bias introduced by rounding values.

    This class should generally be used in combination with a lower bound
    on the integer variables along the lines of ``min((0.2,
    mueff/dimension))``, as it is induced by passing a nonempty
    ``'integer_variables'`` option to `CMAEvolutionStrategy`. Applying
    lower bounds alone is generally more effective than applying integer
    centering alone.

    Class instances have no dynamically changing state variables!

    When integer variable indices are passed to `CMAEvolutionStrategy` via
    the ``'integer_variables'`` `CMAOptions`, an `IntegerCentering` class
    instance is created, and it is called in `CMAEvolutionStrategy.tell`
    before the update of the state variables of CMA, passing mu (genotypic)
    solutions, like ``int_centering(self.pop_sorted[:self.sp.weights.mu],
    self.mean)`` . The call changes the `numpy` arrays of `self.pop_sorted`
    in place. The call tries to access the (phenotypic) bounds as defined
    in ``es.boundary_handler`` from the constructor argument. Hence it is
    expected to fail with a combination of ``bounds`` and
    ``fixed_variables`` set in `CMAOptions`.

    By default, integer centering is applied, e.g. with `fmin2` or
    `CMAEvolutionStrategy`, when the integer variable indices are given
    via the ``'integer_variables'`` option. Since v4.3.0, candidate
    solutions delivered by `CMAEvolutionStrategy.ask` have rounded values
    in all integer variable positions which can be switched off by setting
    `cma.evolution_strategy.round_integer_variables` to `False`.

    A simple example code:

    >>> import cma
    >>>
    >>> def int1_sphere(x):
    ...     return int(x[0] + 0.5 - (x[0] < 0))**2 + 1000 * (sum(xi**2 for xi in x[1:]))

    >>> es = cma.CMAEvolutionStrategy(2 * [2], 0.5,
    ...             {'integer_variables': [0], 'ftarget': 1e-9, 'verbose': -9})
    >>> es = es.optimize(int1_sphere)
    >>> # assert 'ftarget' in es.stop(), es.stop()  # fails sometimes (like 1%-ish)
    >>> # failure is less likely with options ``'tolfun': 0, 'tolfunhist': 0, 'tolflatfitness': 60``

    The following pseudocode shows a direct use case to explicitly see
    how/where integer centering is applied, namely before the state
    variables are updated in `tell`:

    >>> import numpy as np

    >>> # set dimension dependent lower bound on sample standard deviations
    >>> es = cma.CMAEvolutionStrategy(2 * [2], 0.5,
    ...             {'minstd': [0.2, 0], 'popsize_factor': 1.5, 'ftarget': 1e-9, 'verbose': -9})
    >>> # when integer_variables were passed in the above options,
    >>> # we can inactivate the internal integer centering like
    >>> es.integer_centering = lambda *args: args[0] if args else None
    >>> ic = cma.integer_centering.IntegerCentering([0])
    >>> while not es.stop():
    ...     X = es.ask()
    ...     F = [int1_sphere(x) for x in X]
    ...     # change the mu best solutions in place
    ...     _ = ic([X[i] for i in np.argsort(F)[:es.sp.weights.mu]], es.mean)
    ...     es.tell(X, F, check_points=False)  # don't check for changes in X
    ...     # es.logger.add()
    ...     # es.disp()
    >>> # assert 'ftarget' in es.stop(), es.stop()  # fails sometimes (like 1%-ish)

    Details: The default `method2` was used in the below reference and, as
    of version 4.0.0, is activated in `CMAEvolutionStrategy` when
    `integer_variables` are given as option.

    Reference: Marty et al, LB+IC-CMA-ES: Two simple modiﬁcations of
    CMA-ES to handle mixed-integer problems. PPSN 2024.
"""
    def __init__(self, int_indices, method=2, correct_bias=True,
                 repair_into_bounds=True, **kwargs):
        """`int_indices` can also be a `CMAEvolutionStrategy` or `CMAOptions` instance.
        `correct_bias` can be in [0, 1] indicating the fraction of the bias
        that should be corrected (but only up to the mean center).

        `repair_into_bounds` repairs all solutions before centering is
        applied. This does not guaranty solutions to be feasible after
        centering in the case when `np.round` can make a solution
        infeasible.

        Details: When the bias computation and correction is applied only
        to the better half of the population (as by default), the value of
        `correct_bias` must be smaller than 1 to move the mean toward a
        solution that has rank lambda/5 or worse, because ``w_{lambda/5}``
        equals roughly ``1/mu`` (``w_{lambda/3}`` equals roughly
        ``1/mu/2``).

        """
        self.params = dict(kwargs)
        self.params.update({n:v for n, v in locals().items() if n != 'self'})
        self.params.setdefault('bounds_offset', 1e-12)
        '''absolute and relative epsilon offset on the bounds such that
           ``np.round`` should never make a solution infeasible that was
           set on a ``int+-1/2`` bound'''
        if not 0 <= self.params['correct_bias'] <= 1:
            new_value = True if self.params['correct_bias'] else False
            _warnings.warn("correct_bias should be in [0, 1],"
                           " the given value {0} is now interpreted as {1}"
                           .format(self.params['correct_bias'], new_value))
            self.params['correct_bias'] = new_value
        self._int_mask = None
        '''the mask will be set from `int_indices` for coding convenience'''
        self.int_indices = self.get_int_indices(int_indices)
        self._has_bounds = None  # set once in has_bounds
        self._lower_bounds = None  # set once
        self._upper_bounds = None  # set once
        try:
            self.center = getattr(self, 'method' + str(method))
        except AttributeError:
            raise ValueError("`method` argument must be 1 or 2, was: {0}"
                             .format(method))

        # for the record / debugging:
        self._record = False
        self._print = False
        self.mahalanobis0 = []  # only for the record
        self.mahalanobis1 = []  # after modification
        self.last_changes = []
        self.last_changes_iteration = []
        self.centering_on = centering_on

    def bound(self, i):
        """return lower and upper bound on variable `i`, not in use.
        """
        try:
            return self.params['es'].boundary_handler.get_bound(i)
        except Exception:
            return [-np.inf, np.inf]
    def has_bounds(self, which='both'):
        # TODO: make which='lower' and 'upper' work?
        #       It's probably only a minor speedup.
        if which != 'both':
            raise NotImplementedError(
                "which={0} is invalid. Currently, only 'both' bounds can be checked"
                " with `has_bounds`".format(which))
        if self._has_bounds is not None:
            return self._has_bounds
        self._has_bounds = False
        if 'es' not in self.params:
            return False
        try:
            bh = self.params['es'].boundary_handler
        except Exception as e:
            _warnings.warn("``bh = self.params['es'].boundary_handler`` failed with"
                           "\n\n  {0}".format(e))
        try:
            if bh is not None and bh.has_bounds():
                self._has_bounds = True
        except Exception as e:
            _warnings.warn("``bh.has_bounds()`` failed with"
                           "\n\n  {0}".format(e))
        return self._has_bounds
    def lower_bounds(self, dimension):
        """return lower bounds of dimension `dimension` or a scalar.

        `dimension` remains the dimension from the first call unless
        the ``_lower_bounds`` attribute is reset to `None`.
        """
        if self._lower_bounds is None:
            self._lower_bounds = self._set_bounds('lower',
                    np.zeros(dimension) - np.inf)  # np.asarray is slower than np.zeros for dimension >30
        return self._lower_bounds
    def upper_bounds(self, dimension):
        """return upper bounds of dimension `dimension` or a scalar.

        `dimension` remains the dimension from the first call unless
        the ``_upper_bounds`` attribute is reset to `None`.
        """
        if self._upper_bounds is None:
            self._upper_bounds = self._set_bounds('upper',
                    np.zeros(dimension) + np.inf)
        return self._upper_bounds
    def _set_bounds(self, which, bounds):
        """return in place modified `bounds`"""
        assert which in ('lower', 'upper'), which
        if self.has_bounds():
            bounds[:] = self.params['es'].boundary_handler.get_bounds(which, len(bounds))
            if self.params['bounds_offset']:
                # mitigate corner cases such as ``np.round([.5, 1.5]) == [0, 2]``
                # pick variables with value at ..., -0.5, 0.5, 1.5, ...
                idx = np.mod(bounds, 1) == 0.5
                bounds[idx] += (1 if which == 'lower' else
                    -1) * self.params['bounds_offset'] * np.maximum(1, np.abs(bounds[idx]))
        return bounds

    def get_int_indices(self, es_opts_indices=None):
        """determine integer variable indices from es or es.opts or

        a variable index list or self.
        """
        if es_opts_indices is None:  # check self.int_indices and self.params
            return self.int_indices if (
                hasattr(self, 'int_indices') and self.int_indices is not None
                ) else self.params['int_indices']
        if hasattr(es_opts_indices, 'opts'):  # check es
            if 'es' not in self.params:
                self.params['es'] = es_opts_indices
            opts = es_opts_indices.opts
        else:  # assume that arg was an opts dict
            opts = es_opts_indices
        names = ['integer_variables', 'integer_indices', 'int_indices']
        for name in names:
            if name in opts:
                indices = opts[name]
                break  # key found
        else:
            indices = opts
        if 'int_indices' not in self.params:  # can't currently happen
            self.params['int_indices'] = indices
        return indices

    def callback(self, es):
        """change `es.pop_sorted` and update `es.mean` accordingly.

        Not in use, as this must be ideally done in the middle of `tell`,
        that is, neither before nor after `tell`.
        """
        self(es.pop_sorted, es.mean_old)
        es.mean = np.dot(es.sp.weights.positive_weights,
                         es.pop_sorted[0:es.sp.weights.mu])
    @property
    def int_mask(self):
        if self._int_mask is None:
            raise ValueError("dimension is not known yet")
        return self._int_mask

    def __call__(self, solution_list, mean):
        """round values of int-variables in `solution_list` and correct for bias.

        Return `solution_list`, however assignment is deprecated as
        `solution_list` is changed in place and, in particular, because we
        center usually only the first half of the population.

        Variables are only rounded if their rounded value is different from
        the mean. Bias correction is only applied if the rounded value is
        the same as the rounded value of the mean and only towards the
        rounded value of the mean, not away from it.

        None of these changes change the rounded variable value.

        Details: The bias correction could be applied directly to the mean
        after the mean update, as it probably does not have any relevant
        effect on the C update. This would require another call or passing
        the Delta mean in the return value.
    """
        if not self.centering_on:
            return solution_list
        if self._int_mask is None:
            self._int_mask = np.asarray([i in self.int_indices
                                         for i in range(len(mean))])
        if self.params.get('repair_into_bounds', False):
            self.repair(solution_list)  # create a bias towards feasible solutions
            # this does not guaranty that round(x) is also in bounds which
            # may be a good feature when the bounds are chosen suboptimally
        self.center(solution_list, mean)
        if 11 < 3 and self.params.get('repair_into_bounds', False):
            self.repair(solution_list, randomized=False)
        return solution_list

    def method1(self, solution_list, mean):
        """DEPRECATED (experimental and outdated) round values of int-variables in `solution_list` and reduce bias to the mean"""
        m_int = np.round(mean)
        mutated_down = np.mean([np.round(x) < m_int for x in solution_list], axis=0)
        mutated_up = np.mean([np.round(x) > m_int for x in solution_list], axis=0)
        maxmut_ratio = np.where(mutated_down > mutated_up, mutated_down, mutated_up)
        # assert all(maxmut_ratio == np.max([mutated_down, mutated_up], axis=0))  # 3x slower
        self.last_changes = []
        for x in solution_list:
            for i in self.int_indices:
                if (np.round(x[i]) != m_int[i] or (self.params['correct_bias'] and
                    (maxmut_ratio[i] > 0.5 or  # reduce bias
                     np.random.rand() < maxmut_ratio[i] / (1 - mutated_up[i] - mutated_down[i])))):
                     if self._record:
                        self.mahalanobis0.append(self.params['es'].mahalanobis_norm(x - mean))
                     x[i] = np.round(x[i])
                     if self._record:
                        self.mahalanobis1.append(self.params['es'].mahalanobis_norm(x - mean))
                        n0, n1 = self.mahalanobis0[-1], self.mahalanobis1[-1]
                        self.last_changes.append([i, n0, n1])
                        self.last_changes_iteration.append(self.params['es'].countiter)
        return solution_list  # declarative, elements of solution_list have changed in place
    def method2(self, solution_list, mean):
        """center (round) values of int-variables of solutions in `solution_list`.

        Elements of `solution_list` must accept boolean indexing like ``np.arrays``.

        Values are centered iff the centered value differs from the
        centered mean. If `correct_bias`, the introduced bias is amended by
        changing the other (noncentered) coordinates towards their
        int-center (which is the int-center of the mean) too, up to the
        fraction `correct_bias`.

        CAVEAT/TODO: the bias correction currently doesn't check bounds,
        hence it may push feasible solutions out-of-bounds by (i) centering
        single solutions and (ii) shifting coordinates towards their
        centered value during bias correction. In itself and overall, this
        may not be a problem, in particular when `repair` is applied before
        or after calling `method2`.
        """
        repair_int = False
        if self._print:
            mean0 = np.mean(solution_list, axis=0)
        # m_int = np.round(self.repair([mean], randomized=False, copy=True)[0])
        m_int = np.round(mean)
        '''to check whether x_int == m_int and to limit the bias correction move'''
        # Should we check that m_int is in-bounds!? The problem to
        # repair m_int here is that it breaks the ``x_int == m_int``
        # condition. Generally, the mean doesn't need to be feasible and if
        # it is updated only with feasible solutions it becomes feasible.
        dim = len(solution_list[0])
        if self.params['correct_bias']:
            biases = np.zeros(dim)
            '''bias (sum) created from centering solutions'''
            mneg = np.zeros(dim)
            '''sum of possible move to the left for unbiasing'''
            mpos = np.zeros(dim)
            '''sum of possible move to the right for unbiasing'''
        for i, x in enumerate(solution_list):
            x_int = np.round(x)
            ism = (x_int == m_int)
            # The below repairs are only effective (i.e. x_int < lbs is
            # only possible) when x < lbs or the bound has an unreasonable
            # value like 0.25. Then, (i) the below would prevent centering
            # a solution with the boundary domain value at all. (ii)
            # without the below, the bias correct introduces a bias toward
            # the feasible domain (which is not necessarily bad).
            # if self.params.get('repair_into_bounds', False) and self.has_bounds():
            #     # TODO: this breaks the bias correct condition
            #     #       x_int == m_int, which may be fine
            #     ism = np.logical_or(ism, x_int < self.lower_bounds(dim))
            #     ism = np.logical_or(ism, x_int > self.upper_bounds(dim))
            if repair_int:
                self.repair([x_int])
            if self.params['correct_bias']:
                mpos += ism * ((x_int - x) > 0) * (x_int - x)
                mneg += ism * ((x_int - x) < 0) * (x_int - x)
                biases += ~ism * (x_int - x)
            x[~ism * self.int_mask] = x_int[~ism * self.int_mask]  # round off-mean values
        if self.params['correct_bias']:
            # change noncentered variables (those with the int-value of the mean)
            # compare biases with mpos or mneg and set alpha like bias/mpos
            frac = self.params['correct_bias']

            _time_code = False
            if _time_code:  # compare two versions
                import collections, time
                try: self.alpha_timings
                except: self.alpha_timings = collections.defaultdict(float)
                # takes 0.03s (~1.5%) in 200D for 1e4 evaluations
                t0 = time.time()
                alphas = []
                for i, (b, p, n) in enumerate(zip(biases, mpos, mneg)):
                    if not self.int_mask[i]:
                        alphas += [0.]
                    elif b * p < 0:
                        alphas += [-frac * b / p if -frac * b < p else 1.0]
                    elif b * n < 0:
                        alphas += [-frac * b / n if frac * b < -n else 1.0]
                    else:
                        alphas += [0.]
                self.alpha_timings['alpha'] += time.time() - t0
                t0 = time.time()

            # about 4 times faster in 200 D
            alphas2 = np.zeros(len(biases))
            for moves in [mpos, mneg]:
                idx = self.int_mask * biases * moves < 0
                alphas2[idx] = np.minimum(1, -frac * biases[idx] / moves[idx])
                assert moves is mneg or np.random.rand() < 0.98 or (
                    np.all(idx + (biases * mneg < 0) < 2)), (idx, biases, mpos, mneg)

            if _time_code:
                self.alpha_timings['alpha2'] += time.time() - t0
                assert np.all(alphas2 - 1e-11 <= alphas) and (
                       np.all(alphas2 + 1e-11 >= alphas)), (
                            alphas2, alphas, alphas2 - alphas)

            # this does not guaranty feasibility unless above x_int was repaired
            m_int_repaired = (self.repair([np.array(m_int, copy=True)])[0]
                        if repair_int else m_int)
            for x in solution_list:
                x += ((np.round(x) == m_int)
                        * (biases * (m_int_repaired - x) < 0)
                        * alphas2 * (m_int_repaired - x))  # move towards m_int == x_int
        if self._print:  # print remaining biases
            mean1 = np.mean(solution_list, axis=0)
            biases2 = [(i, mean1[i] - mean0[i]) for i in range(len(mean0))
                    if (mean1[i] - mean0[i])**2 > 1e-22]
            if biases2 or not self.params['es'].countiter % 100:
                print(self.params['es'].countiter, alphas2, biases2)
        return solution_list  # declarative, elements of solution_list have changed in place

    def repair(self, solution_list, randomized=True):
        """set values of int-variables of solutions of `solution_list` into bounds.

        Elements of `solution_list` are changed in place after passing them
        through ``np.asarray`` and `solution_list` is changed in place too.

        When ``randomized is True`` sample the value uniformly between the
        bound and the value center (the rounded bound) when the latter is
        feasible.
        """
        if not self.has_bounds():
            return solution_list
        copy = False  # is a constant for the time being
        dim = len(solution_list[0])
        lbs = self.lower_bounds(dim)
        ubs = self.upper_bounds(dim)
        for i, x in enumerate(solution_list[:]):
            x = np.asarray(x)  # this may detach x from solution_list
            # set int-variables of x onto bounds
            islow = np.logical_and(x < lbs, self.int_mask)
            ishigh = np.logical_and(x > ubs, self.int_mask)
            assert np.random.rand() < 0.95 or not np.any(islow * ishigh), (
                i, islow, ishigh)  # rand makes it ~5 times faster
            if np.any(islow):
                if copy:  # FIXME: this is a second copy if asarray made one
                    x = np.array(x, copy=True)
                x[islow] = lbs[islow]
            if np.any(ishigh):
                if copy:  # FIXME: this is a second copy if asarray made one
                    x = np.array(x, copy=True)
                x[ishigh] = ubs[ishigh]
            # move modified variables randomly towards the center
            if randomized:
                x_int = np.round(x)
                idx = np.logical_or(np.logical_and(islow, x_int > x),
                                    np.logical_and(ishigh, x_int < x))
                if np.any(idx):
                    x[idx] += (x_int[idx] - x[idx]) * np.random.rand(sum(idx))
            solution_list[i] = x  # needed when np.asarray above made a copy
        return solution_list



import warnings

def filterwarning(*args, **kwargs):
    """add warning filter but only if not already there (i.e. defined by the user),

    not clear why this works as expected.
    """
    try:
        _is_filtered = any(w[2] == kwargs['category'] for w in warnings.filters)
    except Exception:
        _is_filtered = False
    if not _is_filtered:
        warnings.filterwarnings(*args, **kwargs)

class NeverTestedWarning(UserWarning):
    """Feature has been implemented but never or poorly tested for correct results"""



"""Tools for (lean) experimentation and collecting data.

See also cma.optimization_tools and cma.utilities.math.test*.

TODO: write some running examples as doctests.
"""
from __future__ import division, print_function
import warnings
import os
import shutil
import time
from collections import defaultdict, namedtuple
from ast import literal_eval
import math
import numpy as np  # also to replace 'inf' with np.inf and vice versa
from cma import warnings_and_exceptions as _warnings_and_exceptions


def unique_time_stamp(decimals=2):
    "a unique timestamp up to 1/10^decimals seconds"
    s = '{0:.' + str(decimals) + 'f}' + '.' * (decimals == 0)
    return time.strftime("%Y-%m-%d-%Hh%Mm%Ss") + (
            s.format(time.time()).split('.')[1])

def down_sample(x, y=None, len_=500):
    """return (index, x_down) if y is None, else (x_down, y_down).

    Example: ``plot(*down_sample(mydata))``

    """
    if y is None:
        x, y = np.arange(len(x)), np.asarray(x)
    while len(y) > 2 * len_:
        x = x[::2]
        y = y[::2]
    return x, y

def sp1(data):
    """return the average of finite entries in `data`,

    multiplied by the ratio ``Ndata / Nfinite == len(data) /
    sum(np.isfinite(data))``.

    This measure has been called Q-measure or success performance one, sp1. It
    is computed as the average over all finite entries times the number of all
    entries divided by the number of finite entries. If all entries are finite
    this is the average value.

    Details: the multiplier operates under the assumption that non-finite
    entries contribute the same as the average finite entry and resampling can
    be applied to get a finite entry.
    """
    a = np.asarray(data)
    idx = np.isfinite(a)
    return len(a) * np.mean(a[idx]) / sum(idx) if sum(idx) else np.inf

def copy_file(src, dest):
    """copy src to dest and created dest folder(s) if necessary"""
    if not os.path.exists(src):
        warnings.warn("{0} does not exist (yet), nothing to be copied"
                     .format(src))
        return
    try:
        os.makedirs(os.path.dirname(dest))  # bails when a folder exists
    except Exception:  # error or IOError or OSError or FileExistsError?
        pass
    shutil.copy2(src, dest)

class TimeWarner:
    """context manager to print the time spent iff it exceeds a threshold.

    Usage: ``with TimeWarner('saving'):``, the argument is used only for
    printing the message in case.
"""
    def __init__(self, name, warn_time=2):
        self.name = name
        self.warn_time = warn_time
    def __enter__(self):
        self.t0 = time.time()
        return self
    def __exit__(self, exc_type, exc_value, traceback):
        self.t1 = time.time()
        if self.t1 - self.t0 > self.warn_time:
            print("TimeWarner: {} took {:.1f} seconds".format(self.name, self.t1 - self.t0))

class ClassFromDict(object):
    """set class attributes from a `dict`, see also `cma.utilities.utils.DictClass2`"""
    def __init__(self, dict_):
        self._dict = dict(dict_)
        for key in dict_:
            setattr(self, key, dict_[key])
    @property
    def as_dict(self):
        """collect only original attributes, check out ``.__dict__`` to see also
        the attributes later added.
        """
        return dict([key, getattr(self, key)] for key in self._dict)

class Results(object):
    """a container to communicate (changing) results via disk.

    The ``data`` attribute is a dictionary (`DataDict`) which contains all
    observed data, usually a list for each key. Any new container reads
    previously saved data (with the same name) on initialization.

    Before the data are saved, the old data are backupped under the current
    timestamp.

    Use case example::

        import cma.experimentation

        # create or load 'sweep_data' results
        res = cma.experimentation.Results('sweep_data')

        # save data after each trial
        for dim in dimensions:
            res.backup()  # in case we want fewer backups
            for ...:
                x, es = cma.fmin2(...)
                res.data[dim] += [es.result.evaluations if 'ftarget' in es.result.stop()
                                  else np.inf]
                res.save(backup=False)  # like this we can never lose data


    Details: Saving/loading of nonfinite values with `ast.literat_eval` is
    covered with the ``values_to_string`` parameters.

    Some methods handle nested dictionaries, however it seems to be safer to use
    tuples as keys in a flat dictionary or use different `Results` instances,
    e.g. by dimension and/or popsize, like::

        res = experimentation.Results('sweep_c_dim={}lam={}.format(dimension, popsize)

    In particular ``.data.clean()`` for `float` keys does (currently) not work
    with key tuples.
"""
    def __init__(self,
                 name=None,
                 values_to_string = (np.inf, np.nan, math.nan),
                 verbose=1):
        self.filename = 'results_dict' if name is None else name
        if '.' not in self.filename[-8:]:
            self.filename = self.filename + '.pydict'
        path, name = os.path.split(self.filename)
        name, ext = os.path.splitext(name)
        self._extension = ext
        self.backup_dirname = os.path.join(path, '._' + name + '-backup')
        self._values_to_string_vals = values_to_string
        self.data = DataDict()  # {}
        self.meta_data = dict(info="", )
        if os.path.exists(self.filename):
            self.load()
            if self.data and verbose > 0:
                keys = sorted(self.data)
                print('Results("{3}") loaded {0} data key entries ({1} ... {2})'
                      ' with {4} values'
                      .format(len(self.data), keys[0], keys[-1],
                              self.filename, sum(self.samples().values())))

    def __getattr__(self, name):
        """access ``Results.data`` directly from `Results`"""
        return getattr(self.data, name)

    def _load_json(self):
        """return dictionary from load json file with keys to `int` or `float`, if possible"""
        import json
        filename = self.filename.rsplit('.', 1)[0] + '.json'
        try:
            res = json.load(filename)
        except:
            warnings.warn('json file not loaded, this warning needs to be improved/removed')
            return
        raise NotImplementedError('how do we know whether to convert keys into numbers?')
        for k in sorted(res):  # sort is not needed or effective for now
            for to_number in (int, float):
                try:
                    knew = to_number(k)
                except ValueError:
                    pass
                else:
                    res[knew] = res[k]
                    del res[k]
                    break
        # self.data = DataDict(res)  # this is the idea, then still values need to be fixed
        return res

    def load(self):
        """load data from file, discard current data, see also `update`"""
        if self.data:
            self.backup()
        with open(self.filename, 'rt') as f:
            try:
                self.data = DataDict(literal_eval(f.read()))
            except Exception:
                print("Please check whether inf or nan were used as simple value"
                      " rather than in a sequence as [inf] or [nan].")
                raise
        self._values_to_string(inverse=True)
        self.meta_data = self.data.pop('meta_data', None)
        return self

    def update(self, filename, backup=True):
        """Append data from ``filename`` to ``self.data``.
        
        To update self with a ``data`` `dict` instead a file, call
        ``self.data.update(data)``.

        Details: assumes a flat `dict` structure.
        """
        data = Results(filename).data
        if backup:
            self.backup()
        for key in data:
            self.data[key] += data[key]
        return self

    def save(self, backup=True):
        """save `self.data` to disk"""
        if backup:
            self.backup()
            # try:
            #     with open(self.filename, 'rt') as f:
            #         self.backup(f.read())
            # except IOError: pass
        self._values_to_string()
        if self.data.get('meta_data', None) is None and self.meta_data is not None:
            self.data['meta_data'] = repr(self.meta_data)
        try:
            os.makedirs(os.path.dirname(self.filename))  # bails when a folder exists
        except Exception:  # error or IOError or OSError or FileExistsError?
            pass
        with open(self.filename, 'wt') as f:
            f.write(repr(self.data))
        self.data.pop('meta_data')  # otherwise they get too often in the way
        self._values_to_string(inverse=True)
        return self  # necessary/useful?

    def _value_to_string(self, val, inverse=False):
        """return "correct" value, doesn't work for `nan`"""
        if inverse:
            for val_ in self._values_to_string_vals:
                if repr(val_) == val:
                    return val_
            return val
        if val in self._values_to_string_vals:  # works with nan too
            return repr(val)
        return val
    def _values_to_string(self, inverse=False):
        """replace certain values in lists with strings or vice versa.

        Values to be replaced are defined at instance creation in
        `_values_to_string_vals`.

        Details: this prevents `ast.literal_eval` to bail on, for example,
        ``repr([1, 3, 'inf'])``, as 'inf' is just a string and will not be
        evaluated.
        """

        for a in self._arrays():
            try: len(a), a[0]
            except (TypeError, IndexError):
                try:
                    literal_eval(repr(a))
                except:
                    warnings.warn(
                        "non-sequence found as data value in {0}, which will fail when"
                        "\n  reading back serialized data. A simple fix is to use"
                        "\n  ``[val]`` instead of ``val`` or avoid values which"
                        "\n  `ast.literal_eval` cannot digest, like ``inf``, ``nan``, etc."
                .format(a))
            else:
                for i in range(len(a)):
                    a[i] = self._value_to_string(a[i], inverse)

        for val in self._values_to_string_vals:
            new, old = repr(val), val
            if inverse:
                new, old = old, new
            if old in self.data:
                self.data[new] = self.data[old]
                del self.data[old]

    def _arrays(self, data=None):
        """return a flat list of (references to) all non-dicts in data.

        Traverses recursively down into dictionaries.
        """
        if data is None:
            data = self.data
        if isinstance(data, dict):
            res = []
            for key in data:  # travers down
                if key != 'meta_data':
                    res += self._arrays(data[key])
            return res
        else:  # return leave
            return [data]

    def backup(self):
        """backup saved data by making a file copy"""
        dest = os.path.join(self.backup_dirname,
                            unique_time_stamp(2) + self._extension)
        copy_file(self.filename, dest)
        self.last_backup = dest

    def _backup0(self, data=None):
        """append data to backup file, no easy to read back"""
        if data is None:
            self._values_to_string()
            datastring = repr(self.data)
            self._values_to_string(inverse=True)
        else:
            datastring = repr(data)
        with open(self.backup_dirname + self._extension, 'at') as f:
            f.write(datastring)
            f.write('\n')

class DataDict(defaultdict):
    """A (default) dictionary like ``parameter_value: list_of_measures``,

    e.g. with float parameter or dimension as key and runtimes as value. This
    class is used under the hood in the `Results` class.

    A main functionality is the method `clean`, which joins all entries
    which have almost equal keys. This allows to have a `float` parameter
    as key.

    This class provides simple computations on this kind of data,
    like ``x, y = .xy_arrays() == sorted(keys), sp1(values)``.

    If the dictionary values are not lists, one may get rather unexpected
    results or exceptions.

    Details: this class allows to use `float` values as keys when
    `clean_key` and `set_clean` are used to access the data in the
    `dict`. Inheriting from `defaultdict` with `list` as default value,
    the syntax::

        data = DataDict()
        data[first_key] += [first_data_point]

    without initialization of the key value works perfectly fine.

    Caveat: small values are considered as the same key, even if they are
    close to zero. Either use a different comparison via the `equal`
    keyword parameter, or use ``1 / key_value`` or `log(key_value)``.

    TODO: consider `numpy.allclose` for almost equal comparison?
"""
    def __init__(self, dict_=None):
        """Use ``dict(dict_.data or dict_)``, and `dict_.meta_data` for
        initialization.

        Details: `dict_.meta_data` are assigned as a reference.
        """
        defaultdict.__init__(self, list)
        if dict_ is not None:
            if hasattr(dict_, 'meta_data'):
                self.meta_data = dict_.meta_data
            if hasattr(dict_, 'data'):
                self.update(dict_.data)
            else:
                self.update(dict_)

    def update(self, dict_):
        """update data lists from a `dict` of lists (and only a `dict`)"""
        data = self
        for k in dict_:
            data[k] += dict_[k]

    def xy_arrays(self, agg=sp1, type_=np.asarray):
        """return two arrays ready to be plotted like ``plot(*xy_arrays)``.

        The x-array contains the sorted keys, the y-array contains the
        respectively aggregated values.

        For example to be used like::

            ``plot(*self.xy_arrays())``.

        Parameter `agg` determines the function to aggregate data values, by
        default `sp1` which is the mean corrected for missing data. To show
        dispersion, we can use ``agg=lambda x: np.percentile(x, 10)`` and ``...,
        90)``.
        """
        data = self
        keys = sorted(data)   # [k for k in sorted(self) if np.isscalar(k)]
        return (type_(keys),
                type_([agg(data[k]) for k in keys]))

    def aggregated(self, agg=sp1, relative=False, by=None):
        """WIP return a `dict` with ``(agg(data), number_of_samples)`` per key.

        For example, to get the 10%tile::

            .aggregated(lambda x: np.percentile(x, 10))

        """
        data = self
        def normalize_in_place(res):
            min_ = min(v[0] for v in res.values())
            for k in res:
                res[k] = res[k][0] / min_, res[k][1]
            return res

        keys = sorted(data)  # was: [k for k in sorted(data) if np.isscalar(k)]
        res = {k: (float(agg(data[k])), len(data[k])) for k in keys}
        if relative:
            if by is None:
                normalize_in_place(res)
            else:
                by_values = [k[by] for k in res]
                res_ = {}
                for by_ in sorted(set(by_values)):
                    res_[by_] = normalize_in_place({k: v for k, v in res.items() if k[by] == by_})
                return res_
        return res
    
    def samples(self):
        """return a `dict` with the number of values per key"""
        data = self
        return {k: len(data[k]) for k in sorted(data)}

    def argmin(self, agg=sp1, slack=1.0, slack_index_shift=+1):
        """slack_index_shift can be +-1, looking to the right/left of argmin"""
        x, y = self.xy_arrays(agg)
        idxmin = np.argmin(y)
        if not slack_index_shift:
            return x[idxmin]
        assert slack_index_shift in (-1, 1), slack_index_shift  # quick hack
        i = idxmin + slack_index_shift
        while i >= 0 and i < len(y) and y[i] <= slack * y[idxmin]:
            i += slack_index_shift
        return x[i - slack_index_shift]

    def tests(self, gap=1, method='auto'):
        """return p-values of the `mannwhitneyu` test of entries adjacent with `gap`"""
        import scipy.stats
        data = self
        keys = sorted(data)
        return {(keys[i], keys[i+gap]): float(
                    scipy.stats.mannwhitneyu(
                        data[keys[i]], data[keys[i+gap]],
                        method=method, alternative='two-sided')[1])
                for i in range(len(keys) - gap)}
    def test(self, key, key2=None, method='auto'):
        """return p-value of the `mannwhitneyu` test"""
        import scipy.stats
        data = self
        keys = sorted(data)
        if key not in keys:
            raise ValueError("key {} not in data keys {}".format(key, keys))
        if key2 is None:
            idx = keys.index(key) + 1
            if idx >= len(keys):
                raise ValueError("key {} is the last key in data {}".format(key, keys))
            key2 = keys[idx]
        return {(key, key2): float(
                    scipy.stats.mannwhitneyu(
                        data[key], data[key2],
                        method=method, alternative='two-sided')[1])}

    def clean(self, equal=lambda x, y: x - 1e-6 < y < x + 1e-6):
        """merge keys which have almost the same value"""
        for key in list(self.keys()):
            if key not in self:  # self.keys() changes in the process
                continue
            self.clean_key(key, equal=equal)
        return self

    def clean_key(self, key, equal=lambda x, y: x - 1e-6 < y < x + 1e-6):
        """set similar key values all to be `key`, return `key`.

        Use method `set_clean` to access and change the clean-key
        dictionary *value* more conveniently.
        """
        meta_data = self.pop('meta_data', None)
        while self._near_key(key, equal) is not key:
            k = self._near_key(key, equal)
            assert k != key
            self[key] += self[k]  # def join_(a, b): return list(a) + list(b) could become input
            del self[k]  # del a superfluous reference
        if meta_data is not None:
            self['meta_data'] = meta_data
        return key

    def get_near(self, key, equal=lambda x, y: x - 1e-6 < y < x + 1e-6):
        """get the merged values list of all nearby keys.

        Caveat: the returned value is a new list

        :See also: `clean`, `set_clean`.
        """
        res = [] if key not in self else list(self[key])  # prevent adding the key
        done_keys = [key]
        while self._near_key(key, equal, done_keys) is not key:
            k = self._near_key(key, equal, done_keys)
            done_keys += [k]
            res += self[k]
        return res

    def set_clean(self, key):
        """join all entries with similar `key` and return the new value,
        a joined list of all respective values.

        This is the same as `clean_key` which however returns the new key, not
        the values.

        Example::

            data.set_clean(key) += [new_data_point]

            # same as
            data[data.clean_key(key)] += [new_data_point]

            # or more explicite, however with a different order of the data
            data[key] += [new_data_point]
            data.clean_key(key)  # joins data, however in the "wrong" order

            # similar as
            data[key] += [new_data_point]
            data.clean()  # cleans *all* keys

        """
        self.clean_key(key)
        return self[key]  # same as self[self.clean_key(key)]

    def _near_key(self, key, equal=lambda x, y: x - 1e-6 < y < x + 1e-6,
                  exclude=None):
        """return a key in self which is ``equal`` to ``key`` and otherwise ``key``.
        """
        if exclude is None:
            exclude = []
        for k in sorted(self.keys()):  # sorted doesn't work with float and str
            try:  # prevent type error from equal(float, str)
                if equal(k, key) and k != key and k not in exclude:
                    return k
            except TypeError:
                pass
        return key

    @property
    def successes(self):
        """return a class instance with attributes `x` (i.e. keys), `n`,
        `nsucc`, and `rate` as arrays.

        TODO: consider using cma.utilities.utils.DictClass2 instead namedtuple?
        """
        keys = [k for k in sorted(self) if np.isscalar(k)]
        nsucc = np.asarray([sum(np.isfinite(self[k])) for k in keys])
        n = np.asarray([len(self[k]) for k in keys])
        try:
            Successes = namedtuple('Successes',
                                   ['x', 'nsucc', 'n', 'rate'])
            res = Successes(np.asarray(keys), nsucc, n, nsucc / n)
        except:
            res = ClassFromDict(
                {'x': np.asarray(keys),
                 'nsucc': nsucc,
                 'n': n,
                 'rate': nsucc / n,
                 })
        return res

    def percentile(self, prctile, agg=sp1, samples=100):
        """TODO:review percentile based on bootstrapping"""
        raise NotImplementedError
        keys = [k for k in sorted(self) if np.isscalar(k)]
        res = []
        for k in keys:
            bstrapped = []
            for i in range(samples):
                idx = np.random.randint(len(self[k]))
                data = self[k][idx]
                bstrapped.append(agg(data))
            res.append(np.percentile(bstrapped, prctile))
        return np.asarray(keys), np.asarray(res)

    def __repr__(self):
        return repr(dict(self))

class ResultsPandas:
    """WIP ad hoc code for writing (changing) results in a `pandas.DataFrame` and to disk,

    where "results" refers to end-results of experiment repetitions rather
    than the traces of single runs.

    Caveat: this is an ad hoc implementation, some interfaces may be incomplete,
    interface details are still versatile and in flux, some recent changes may be
    faulty.

    Main features, similar to `Results`, are

    * Intermediate savings of results which consequently can be loaded from a
      different shell while the experiment is running.
    * Backup under the current timestamp before each saving into a
      ``'backups-name'`` folder (optional but default).
    * Similar float values can be "equalized" for correct data aggregation, see
      the `check_close_values` and `equalize_close_values` methods. This uses
      `np.isclose` which considers 1e-8 to be close to 1e-9 and 1+1e-5 close to
      1+1e-6.

    Compared to `Results`, this class is useful to store more information for
    each run, like the termination condition or a final condition number or
    constraints violations or meta parameter information. (In contrast, with
    `Results` the workaround for catching the final condition would writing a
    nonfinite entry when the target was not reached.)

    A guiding code example::

        import cma.experimentation

        res = cma.experimentation.ResultsPandas('some-name')  # reloads data to append/continue
        for dim in dimensions:
            es = cma.CMA(dim * [2], 1, {'verbose': -9})
            # a tracker in case we want to track, say, a minimum over the trace as result
            es.optimize(cma.ff.rosen, callback=my_result_tracker)
            if es.opts.get('verbose') == -9:  # do not save data while testing the setup
                res.append([es.N, es.popsize,
                            es.result.evaluations,
                            1 if 'ftarget' in es.stop() else 0,
                            repr(es.stop),
                            es.condition_number,
                            my_result_tracker.my_val_of_interest],
                            colums=['dimension', 'popsize',
                                    'evaluations',
                                    'targethit',
                                    'stopcondition',
                                    'conditionnumber',
                                    'myvalue'])
                res.save()

        # load the data:
        res = cma.experimentation.ResultsPandas('some-name')
        print(res.summary)  # summary statistics of columns in a dict
        res.df  # the pandas data frame

    Notes: ``self.df.drop(...)`` could be used to return a data frame with some
    entried dropped.
"""
    def __init__(self, filepathname, format='.csv'):
        """load data when `filepathname` exists.

        ``format='.csv'`` is human readible, however ``'.feather'`` is much more
        performant, ``'.parquet'`` should work too.
        """
        try:
            import pandas as pd  # noqa: F811
        except ImportError:
            warnings.warn("Please 'pip install pandas' to use the `ResultsPandas` class")
            raise

        self.name = filepathname
        self._extension = format  # feather format seems best performing binary 
                                  # see https://towardsdatascience.com/the-best-format-to-save-pandas-data-414dca023e0d
        # was: self.backup_dirname = '._{0}_backups'.format(self.name)
        if format == '.feather':
            try:
                import pyarrow
            except ImportError:
                warnings.warn("to use the .feather format: 'pip install pyarrow'"
                            '\nusing .csv for the time being')
                self._extension = '.csv'
        p, n = os.path.split(self.name)
        self.backup_dirname = os.path.join(p, 'backups-' + n)
        if os.path.exists(self.name + self._extension):
            self.load()
        else:
            self.df = pd.DataFrame()
        warnings.warn("The ResultsPandas class is versatile and has not be thoroughly tested",
                      category=_warnings_and_exceptions.NeverTestedWarning)
    def __getattr__(self, name):
        """access the `DataFrame` ``Results.df`` directly from `Results`"""
        return getattr(self.df, name)

    @property
    def summary(self):
        """return number of finite entries, number of different values, and

        (min, max) for each column.
        """
        res = self
        samples = {}
        for name in res.df.columns:
            min_, max_ = res.df[name].dropna().min(), res.df[name].dropna().max()
            if isinstance(min_, (int, np.integer)) and isinstance(max_, (int, np.integer)):
                min_, max_ = int(min_), int(max_)
            else:
                try:
                    min_, max_ = float(min_), float(max_)
                except Exception:
                    pass
            try:
                n_valid = len(res.df[name].dropna())
            except Exception:
                n_valid = int(sum(v not in ('nan', '') for v in res.df[name]))
            samples[name] = [n_valid,
                             len(set(res.df[name])),
                              (min_, max_)]
        return samples

    def save(self, backup=True, time_s=2, **kwargs):
        """save data, warn when saving takes more than `time_s` seconds.

        `kwargs` are passed to the saving method of the `pandas` data frame.

        Details: `save` is in essence just a shortcut for::

            self.backup()
            self.df.to_feather(self.name + self.extension)  # assuming .extension == '.feather'

        TODO: generalize by passing the `DataFrame` method name for saving, like
        ``.save('to_feather')``? Annoyingly, pandas does not add a proper
        extension by default.
        """
        backup and self.backup()  # TODO: saves wrong format when self._extension changed
        try:
            os.makedirs(os.path.dirname(self.name + self._extension))  # bails when a folder exists
        except Exception:  # error or IOError or OSError or FileExistsError?
            pass
        with TimeWarner('save', time_s):
            if self._extension == '.csv':
                kwargs.setdefault('index', False)
                self.df.to_csv(self.name + self._extension, **kwargs)
            elif self._extension == '.parquet':
                self.df.to_parquet(self.name + self._extension, **kwargs)
            else:
                if self._extension != '.feather':
                    warnings.warn("'{0}' extension not recognized, saving in feather format"
                                  .format(self._extension))
                self.df.to_feather(self.name + '.feather', **kwargs)
                # see https://towardsdatascience.com/the-best-format-to-save-pandas-data-414dca023e0d
    def load(self, polish=False):
        import pandas as pd  # noqa: F811
        with TimeWarner('ResultsPandas.load'):
            if self._extension == '.csv':
                self.df = pd.read_csv(self.name + self._extension)
            elif self._extention == '.parqet':
                self.df = pd.read_parquet(self.name + self._extension)
            else:
                self.df = pd.read_feather(self.name + self._extension)
        # self.equalize_close_values()
        # polish and self._polish()
        # self.check()
        return self
    def equalize_close_values(self, column):
        df = self.df
        s = sorted(set(df[column]))
        for i, t in enumerate(np.isclose(s[:-1], s[1:])):
            if t:  # values s[i] and s[i+1] are (too) close -> copy s[i] value to rows with s[i+1] value
                df.loc[getattr(df, column) == s[i+1], column] = s[i]
                # df.loc[...] is a view df[...] is a copy    
    def _polish(self, column):
        """a quick hack"""
        self.df['success'] = ['callback' in s for s in self.df.stop]  # slower: self.df.stop.apply(lambda s: 'callback' in s)
        # insert best column (best by dimension)
        idx = self.df_succ.groupby(['dimension']).evals.idxmin()  # don't know how to make .median() work
        self.df['best_' + column] = [getattr(self.df.loc[idx[dim]], column) if dim in idx else np.nan
                                   for dim in self.df.dimension]
    def check_close_values(self, column):
        s = sorted(set(getattr(self.df, column)))
        if any(np.isclose(s[:-1], s[1:])):
            warnings.warn("Found ambiguous double values in column {0} in {1}."
                          "Call .equalize_close_values() and .save() to fix."
                          .format(column, s))

    def append(self, data, columns, kwargs_df={}, kwargs_concat={}):
        """append a single data row"""
        # self.df = self.df.append(*args, **kwargs)
        self.extend([data], columns, kwargs_df, kwargs_concat)
    def extend(self, data, columns, kwargs_df={}, kwargs_concat={}):
        """extend frame by data which is a sequence of rows"""
        import pandas as pd  # noqa: F811

        kwargs_concat.setdefault('ignore_index', True)
        self.df = pd.concat([self.df,
            pd.DataFrame.from_records(data, columns=columns, **kwargs_df)], **kwargs_concat)
    def drop(self, *args, **kwargs):
        """call ``self.df.drop`` and reassign result"""
        raise NotImplementedError
        self.df = self.df.drop(*args, **kwargs)
    def reset_index(self):
        """caveat: the original index is lost which may be undesirable"""
        self.df.reset_index(drop=True, inplace=True)
    def column_values(self, column, by='dimension'):
        """a sorted list of values in the column of name `column`."""
        df, res = self.df, {}
        if by:
            for dimension in sorted(set(df[by])):
                res[dimension] = sorted(set(df[df[by] == dimension][column]))
        else:
            res = sorted(set(df[column]))
        return res              
    @property
    def failure_indices(self, column):
        """TODO: revise such that we have a boolean failed_setting column"""
        failed = [d.Index for d in self.df.itertuples() if 'callback' not in d.stop]
        def failed_parameter(d, failed_indices):
            for i in failed_indices:
                if (d.dimension, getattr(d, column)) == (self.df.loc[i].dimension, getattr(self.df.loc[i], column)):
                    return True
            return False
        return [d.Index for d in self.df.itertuples() if failed_parameter(d, failed)]
    def print_checks(self, column):
        """do not use `iterrows` but `itertuples`
        https://stackoverflow.com/questions/16476924/how-to-iterate-over-rows-in-a-dataframe-in-pandas
        """
        df = self.df
        print("{0} data, {1} failures, failure_indices = {2}\n".format(len(df), len(self.failure_indices), self.failure_indices),
              "min {0} fraction = {1:.4}".format(column, min(np.exp(np.diff(np.log(sorted(set(df[column]))))))),
             )

    def backup(self):
        """backup saved data by making a file copy"""
        dest = os.path.join(self.backup_dirname,
                            unique_time_stamp(2) + self._extension)
        copy_file(self.name + self._extension, dest)
        self.last_backup = dest

del division, print_function



"""Basic implementations of some optimization algorithms namely

- `purecma.fmin` and `purecma.CMAES`: a basic and readible CMA-ES implementation
  where the core algorithm is coded in under 40 lines (in the method `CMAES.tell`)
- `compact_ga.CompactGA`: compact GA which is the most logical abstraction of an
  genetic algorithm as an estimation of distribution algorithm

"""
